#! /usr/bin/env python3
"""
Translate NSO Device config to MDD OpenConfig

This script will pull a device's configuration from an NSO server, convert the NED structured configuration to
MDD OpenConfig, save the NSO configuration to a file named {device_name}_ned_configuration_network_instances.json,
save the NSO device configuration minus parts replaced by OpenConfig to a file named
{device_name}_ned_configuration_remaining_network_instances.json, and save the MDD OpenConfig configuration to a file
named {nso_device}_openconfig_network_instances.json.

The script requires the following environment variables:
NSO_URL - URL for the NSO server
NSO_USERNAME
NSO_PASSWORD
NSO_DEVICE - NSO device name for configuration translation
TEST - True or False. True enables sending the OpenConfig to the NSO server after generation
"""

import copy
import ipaddress
from functools import cmp_to_key
from importlib.util import find_spec

if (find_spec("package_nso_to_oc") is not None):
    from package_nso_to_oc import common
    from package_nso_to_oc.xe import common_xe
else:
    import common
    from xe import common_xe

ospf_network_types = {
    "broadcast": "BROADCAST_NETWORK",
    "point-to-point": "POINT_TO_POINT_NETWORK",
    "non-broadcast": "NON_BROADCAST_NETWORK"
}
xe_routes_notes = []

def configure_xe_static_routes(net_inst, vrf_forwarding_list, config_leftover, network_instances_notes):
    instance_name = net_inst["openconfig-network-instance:name"]
    net_protocols = net_inst["openconfig-network-instance:protocols"]["openconfig-network-instance:protocol"]
    static_route = get_static_protocol(net_protocols)["openconfig-network-instance:static-routes"][
        "openconfig-network-instance:static"]
    route_leftover = config_leftover.get("tailf-ned-cisco-ios:ip", {"route": {}}).get("route", {})

    if instance_name == "default":
        ip_forwarding_list_leftover = route_leftover.get(common_xe.IP_FORWARDING_LIST, [])
        intf_list_leftover = route_leftover.get(common_xe.INTF_LIST, [])
        ip_intf_forwarding_list_leftover = route_leftover.get(common_xe.IP_INTF_FORWARDING_LIST, [])
    else:
        vrf_forwarding_list_leftover = route_leftover.get("vrf", [])
        vrf_index = vrf_forwarding_list["vrf-index"]
        vrf_forwarding_leftover = vrf_forwarding_list_leftover[vrf_index] if len(vrf_forwarding_list_leftover) > vrf_index else {}
        ip_forwarding_list_leftover = vrf_forwarding_leftover.get(common_xe.IP_FORWARDING_LIST, [])
        intf_list_leftover = vrf_forwarding_leftover.get(common_xe.INTF_LIST, [])
        ip_intf_forwarding_list_leftover = vrf_forwarding_leftover.get(common_xe.IP_INTF_FORWARDING_LIST, [])

    for index, route_forwarding in enumerate(vrf_forwarding_list[common_xe.IP_FORWARDING_LIST]):
        process_static_routes(route_forwarding, static_route, ip_forwarding_list_leftover, index)
    for index, route_forwarding in enumerate(vrf_forwarding_list[common_xe.INTF_LIST]):
        process_static_routes(route_forwarding, static_route, intf_list_leftover, index)
    for index, route_forwarding in enumerate(vrf_forwarding_list[common_xe.IP_INTF_FORWARDING_LIST]):
        process_static_routes(route_forwarding, static_route, ip_intf_forwarding_list_leftover, index)
    
    network_instances_notes += xe_routes_notes

def process_static_routes(route_forwarding, static_route, forwarding_list_leftover, index):
    new_static_route = configure_static_route(route_forwarding)

    if new_static_route:
        static_route.append(new_static_route)

        if len(forwarding_list_leftover) >= index:
            forwarding_list_leftover[index] = None

def get_static_protocol(net_protocols):   
    static_protocol = {
        "openconfig-network-instance:identifier": "STATIC",
        "openconfig-network-instance:name": "DEFAULT"
    }
    static_list_template = {"openconfig-network-instance:static": []}

    for net_protocol in net_protocols:
        if net_protocol.get("openconfig-network-instance:identifier", "") == "STATIC":
            if not "openconfig-network-instance:config" in net_protocol:
                net_protocol["openconfig-network-instance:config"] = copy.deepcopy(static_protocol)
            if not "openconfig-network-instance:static-routes" in net_protocol:
                net_protocol["openconfig-network-instance:static-routes"] = copy.deepcopy(static_list_template)
            
            return net_protocol

    static_protocol["openconfig-network-instance:config"] = copy.deepcopy(static_protocol)
    static_protocol["openconfig-network-instance:static-routes"] = copy.deepcopy(static_list_template)
    net_protocols.append(static_protocol)

    return static_protocol

def configure_static_route(route_forwarding):
    if (not "prefix" in route_forwarding or not "mask" in route_forwarding 
        or (not "forwarding-address" in route_forwarding and not "interface" in route_forwarding)):
        
        raise ValueError('Required prefix, mask, or forwarding-address and interface properties is missing.')

    ip_and_mask = ipaddress.ip_network(f"{route_forwarding['prefix']}/{route_forwarding['mask']}")
    new_static_route = {
        "openconfig-network-instance:prefix": ip_and_mask.with_prefixlen,
        "openconfig-network-instance:config": {
            "openconfig-network-instance:prefix": ip_and_mask.with_prefixlen
        },
        "openconfig-network-instance:next-hops": {
            "openconfig-network-instance:next-hop": []
        }
    }

    if "name" in route_forwarding:
        new_static_route["openconfig-network-instance:config"]["openconfig-network-instance:description"] = route_forwarding["name"]
    if "tag" in route_forwarding:
        new_static_route["openconfig-network-instance:config"]["openconfig-network-instance:set-tag"] = route_forwarding["tag"]

    new_index = configure_next_hop_index(route_forwarding)
    new_static_route["openconfig-network-instance:next-hops"]["openconfig-network-instance:next-hop"].append(new_index)

    return new_static_route

def configure_next_hop_index(route_forwarding):
    new_index = {}

    if {"interface", "forwarding-address"} <= route_forwarding.keys():
        (intf_type, intf_num_full) = common.get_interface_type_number_and_subinterface(route_forwarding["interface"])
        (intf_num, sub_intf_num) = common.get_interface_number_split(intf_num_full)
        new_index["openconfig-network-instance:index"] = f"{intf_type}{intf_num}"
        new_index["openconfig-network-instance:config"] = {
            "openconfig-network-instance:index": f"{intf_type}{intf_num}",
            "openconfig-network-instance:next-hop": route_forwarding["forwarding-address"]
        }
        new_index["openconfig-network-instance:interface-ref"] = {
            "openconfig-network-instance:config": {
                "openconfig-network-instance:interface": f"{intf_type}{intf_num}"
            }
        }
        if (intf_type != "Tunnel") and (intf_type != "Vlan"):
            new_index["openconfig-network-instance:interface-ref"]["openconfig-network-instance:config"][
                "openconfig-network-instance:subinterface"] = sub_intf_num
    elif "forwarding-address" in route_forwarding:
        new_index["openconfig-network-instance:index"] = route_forwarding["forwarding-address"]
        new_index["openconfig-network-instance:config"] = {
            "openconfig-network-instance:index": route_forwarding["forwarding-address"],
            "openconfig-network-instance:next-hop": route_forwarding["forwarding-address"]
        }
    elif "interface" in route_forwarding:
        if route_forwarding["interface"] == "Null0":
            new_index["openconfig-network-instance:index"] = "DROP"
            new_index["openconfig-network-instance:config"] = {
                "openconfig-network-instance:index": "DROP",
                "openconfig-network-instance:next-hop": "DROP"
            }
        elif route_forwarding["interface"] == "dhcp":
            new_index["openconfig-network-instance:index"] = "DHCP"
            new_index["openconfig-network-instance:config"] = {
                "openconfig-network-instance:index": "DHCP",
                "openconfig-network-instance:next-hop": "DHCP"
            }
        else:
            (intf_type, intf_num_full) = common.get_interface_type_number_and_subinterface(route_forwarding["interface"])
            (intf_num, sub_intf_num) = common.get_interface_number_split(intf_num_full)
            new_index["openconfig-network-instance:index"] = f"{intf_type}{intf_num}"
            new_index["openconfig-network-instance:config"] = {
                "openconfig-network-instance:index": f"{intf_type}{intf_num}",
                "openconfig-network-instance:next-hop": "LOCAL_LINK",
                "openconfig-local-routing-ext:dhcp-learned": "ENABLE" if "dhcp" in route_forwarding else "DISABLE"
            }
            new_index["openconfig-network-instance:interface-ref"] = {
                "openconfig-network-instance:config": {
                    "openconfig-network-instance:interface": f"{intf_type}{intf_num}"
                }
            }
            if (intf_type != "Tunnel") and (intf_type != "Vlan"):
                new_index["openconfig-network-instance:interface-ref"]["openconfig-network-instance:config"][
                    "openconfig-network-instance:subinterface"] = sub_intf_num

    if "metric" in route_forwarding:
        new_index["openconfig-network-instance:config"]["openconfig-network-instance:metric"] = route_forwarding["metric"]

    return new_index



#! /usr/bin/env python3

import os
import json
import urllib3
import re
from pathlib import Path, os as path_os
from typing import Tuple


if os.environ.get("NSO_URL", False) and os.environ.get("NSO_NED_FILE", False):
    print("environment variable NSO_URL or NSO_NED_FILE must be set: not both")
    exit()
elif not os.environ.get("NSO_URL", False) and not os.environ.get("NSO_NED_FILE", False):
    print("environment variable NSO_URL or NSO_NED_FILE must be set")
    exit()

# Different device OS
XE = "xe"
XR = "xr"

# In case Host OS can't resolve port name to number for ACLs
port_name_number_mapping = {"netbios-ss": 139,
                            "non500-isakmp": 4500,
                            "lpd": 515}

def nso_get_device_config(nso_api_url: str, username: str, password: str, device: str) -> dict:
    """
    Get device configuration from NSO. Return configuration as python dict.
    :param nso_api_url: str
    :param username: str
    :param password: str
    :param device: str
    :return: NSO Device configuration
    """
    url = f"{nso_api_url}/restconf/data/tailf-ncs:devices/device={device}/config"
    req = urllib3.PoolManager(cert_reqs='CERT_NONE')
    headers = urllib3.make_headers(basic_auth=f"{username}:{password}")
    headers.update({"Content-Type": "application/yang-data+json",
                    "Accept": "application/yang-data+json"})
    configuration_result = req.request("GET", url, headers=headers)
    config_before_string = configuration_result.data.decode()
    return json.loads(config_before_string)["tailf-ncs:config"]


def xe_system_get_interface_ip_address(config_before: dict) -> dict:
    """
    Receives an NSO xe configuration and return a dict of interface names to IP addresses.
    E.g, {"GigabitEthernet6": "172.60.1.2"}
    :param config_before: dict
    :return: interface_ip_name dict
    """
    interface_ip_name = {}
    for if_type in config_before["tailf-ned-cisco-ios:interface"]:
        temp_dict = {}
        if if_type == "Port-channel-subinterface":
            for number in config_before["tailf-ned-cisco-ios:interface"]["Port-channel-subinterface"]["Port-channel"]:
                if number.get("ip", {}).get("address", {}).get("primary", {}).get("address"):
                    temp_dict.update({f"Port-channel{number['name']}": f"{number.get('ip', {}).get('address', {}).get('primary', {}).get('address')}"})
            interface_ip_name.update(temp_dict)
        elif if_type == "LISP-subinterface":
            for number in config_before["tailf-ned-cisco-ios:interface"]["LISP-subinterface"]["LISP"]:
                if number.get("ip", {}).get("address", {}).get("primary", {}).get("address"):
                    temp_dict.update({
                                         f"LISP{number['name']}": f"{number.get('ip', {}).get('address', {}).get('primary', {}).get('address')}"})
            interface_ip_name.update(temp_dict)
        else:
            for number in config_before["tailf-ned-cisco-ios:interface"][if_type]:
                if number.get("ip", {}).get("address", {}).get("primary", {}).get("address"):
                    temp_dict.update({f"{if_type}{number['name']}": f"{number.get('ip', {}).get('address', {}).get('primary', {}).get('address')}"})
            interface_ip_name.update(temp_dict)
    return interface_ip_name


def test_nso_program_oc(nso_api_url: str, username: str, password: str, device: str, oc_config: dict) -> None:
    """
    Send translated Openconfig device configuration to NSO
    :param nso_api_url: str
    :param username: str
    :param password: str
    :param device: str
    :param oc_config: dict
    :return: None
    """
    url = f"{nso_api_url}/restconf/data/tailf-ncs:devices/device={device}/mdd:openconfig"
    req = urllib3.PoolManager(cert_reqs='CERT_NONE')
    headers = urllib3.make_headers(basic_auth=f"{username}:{password}")
    headers.update({"Content-Type": "application/yang-data+json",
                    "Accept": "application/yang-data+json"})
    oc = {"mdd:openconfig": {}}
    oc["mdd:openconfig"].update(oc_config)
    body = json.dumps(oc)
    oc_result = req.request("PATCH", url, headers=headers, body=body)
    if oc_result.status != 204:
        if hasattr(oc_result, 'data'):
            raise Exception(f"Error in input payload reported by NSO: {oc_result.data}")
        else:
            raise Exception(f"Error in input payload reported by NSO")

def print_and_test_configs(device_name, config_before_dict, config_leftover_dict, oc, config_name, 
    config_remaining_name, oc_name, translation_notes = []):
    (nso_api_url, nso_username, nso_password) = get_nso_creds()
    nso_device = os.environ.get("NSO_DEVICE", device_name)
    test = os.environ.get("TEST", "False")

    # Determine the project root dir, where we will create our output_data dir (if it doesn't exist).
    # output_data_dir is meant to contain data/config files that we don't want in version control.
    # project_path = str(Path(__file__).resolve().parents[1])
    project_path = os.getcwd()
    output_data_dir = f"{project_path}{path_os.sep}output_data{path_os.sep}"
    Path(output_data_dir).mkdir(parents=True, exist_ok=True)

    print(json.dumps(oc, indent=4))
    with open(f"{output_data_dir}{nso_device}{config_name}.json", "w") as b:
        b.write(json.dumps(config_before_dict, indent=4))
    with open(f"{output_data_dir}{nso_device}{config_remaining_name}.json", "w") as a:
        a.write(json.dumps(config_leftover_dict, indent=4))
    with open(f"{output_data_dir}{nso_device}{oc_name}.json", "w") as o:
        o.write(json.dumps(oc, indent=4))

    if len(translation_notes) > 0:
        # Only print to file, if actual notes exist.
        with open(f"{output_data_dir}{nso_device}{config_name}_notes.txt", "w") as o:
            # We run it through a map, just in case an element in our list of notes contain non-string type.
            # Otherwise, we risk an error when joining.
            o.write("\n\n".join(map(lambda note: str(note), translation_notes)))

    if test == "True":
        test_nso_program_oc(nso_api_url, nso_username, nso_password, nso_device, oc["mdd:openconfig"] if "mdd:openconfig" in oc else oc)

def get_nso_creds():
    nso_api_url = os.environ.get("NSO_URL")
    nso_username = os.environ.get("NSO_USERNAME", "ubuntu")
    nso_password = os.environ.get("NSO_PASSWORD", "admin")

    return (nso_api_url, nso_username, nso_password)

def get_interface_type_number_and_subinterface(interface: str) -> Tuple[str, str]:
    """
    Receive full interface name. Returns interface type and number.
    :param interface: full interface name
    :return: tuple of interface type, interface number.subinterface number
    """
    rt = re.search(r'\D+', interface)
    interface_name = rt.group(0)
    rn = re.search(r'[0-9]+(\/[0-9]+)*(\.[0-9]+)*', interface)
    interface_number = rn.group(0)

    return interface_name, interface_number

def get_index_or_default(obj, index, default = {}):
    try:
        return obj[index]
    except:
        return default

def get_interface_number_split(interface_number: str) -> Tuple[int, int]:
    number_split = interface_number.split('.')

    return tuple(number_split) if len(number_split) > 1 else (number_split[0], 0)



#! /usr/bin/env python3
"""
Translate NSO Device config to MDD OpenConfig

This script will pull a device's configuration from an NSO server, convert the NED structured configuration to
MDD OpenConfig, save the NSO configuration to a file named {device_name}_ned_configuration_network_instances.json,
save the NSO device configuration minus parts replaced by OpenConfig to a file named
{device_name}_ned_configuration_remaining_network_instances.json, and save the MDD OpenConfig configuration to a file
named {nso_device}_openconfig_network_instances.json.

The script requires the following environment variables:
NSO_URL - URL for the NSO server
NSO_USERNAME
NSO_PASSWORD
NSO_DEVICE - NSO device name for configuration translation
TEST - True or False. True enables sending the OpenConfig to the NSO server after generation
"""

import sys
from pathlib import Path
from importlib.util import find_spec
import copy

network_instances_notes = []

openconfig_network_instances = {
    "openconfig-network-instance:network-instances": {
        "openconfig-network-instance:network-instance": [
            {
                "openconfig-network-instance:name": "default",
                "openconfig-network-instance:config": {
                    "openconfig-network-instance:name": "default",
                    "openconfig-network-instance:type": "DEFAULT_INSTANCE",
                    "openconfig-network-instance:enabled": "true"
                },
                "openconfig-network-instance:protocols": {"openconfig-network-instance:protocol": []},
                "openconfig-network-instance:interfaces": {"openconfig-network-instance:interface": []},
                "openconfig-network-instance:vlans": {}
            }
        ]
    }
}

def xe_network_instances(config_before: dict, config_leftover: dict) -> None:
    """
    Translates NSO XE NED to MDD OpenConfig Network Instances
    """
    if config_before.get("tailf-ned-cisco-ios:vrf", {}).get("definition"):
        for vrf_index, vrf in enumerate(config_before.get("tailf-ned-cisco-ios:vrf", {}).get("definition")):
            if vrf.get("address-family"):
                address_families = []
                for key in vrf.get("address-family").keys():
                    if key == "ipv4":
                        address_families.append("openconfig-types:IPV4")
                    if key == "ipv6":
                        address_families.append("openconfig-types:IPV6")
                temp_vrf = {
                    "openconfig-network-instance:name": vrf["name"],
                    "openconfig-network-instance:config": {
                        "openconfig-network-instance:name": vrf["name"],
                        "openconfig-network-instance:type": "L3VRF",
                        "openconfig-network-instance:enabled": "true",
                        "openconfig-network-instance:enabled-address-families": address_families
                    },
                    "openconfig-network-instance:protocols": {"openconfig-network-instance:protocol": []},
                    "openconfig-network-instance:interfaces": {"openconfig-network-instance:interface": []}
                }
                process_rd_rt(temp_vrf, vrf, vrf_index, config_leftover)

                del config_leftover["tailf-ned-cisco-ios:vrf"]["definition"][vrf_index]["address-family"]
            openconfig_network_instances["openconfig-network-instance:network-instances"][
                "openconfig-network-instance:network-instance"].append(temp_vrf)

    interfaces_by_vrf = get_interfaces_by_vrf(config_before)
    route_forwarding_list_by_vrf = get_route_forwarding_list_by_vrf(config_before)
    configure_network_instances(config_before, config_leftover, interfaces_by_vrf, route_forwarding_list_by_vrf)

    cleanup_null_ospf_leftovers(config_leftover)
    cleanup_null_static_route_leftovers(config_leftover)

def get_interfaces_by_vrf(config_before):
    interfaces_by_vrf = {}
    interfaces = config_before.get("tailf-ned-cisco-ios:interface", {})
    for interface_type, interface_list in interfaces.items():
        if interface_type == "Port-channel-subinterface":
            interface_type = "Port-channel"
            interface_list = interface_list[interface_type]

        for interface in interface_list:
            if (not "ip" in interface or not "address" in interface["ip"] 
                or not "primary" in interface["ip"]["address"] or not "address" in interface["ip"]["address"]["primary"]):
                continue

            interface_copy = copy.deepcopy(interface)
            interface_copy["type"] = interface_type
            # Ensure we get a string type
            interface_copy["name"] = str(interface_copy["name"])

            if "vrf" in interface_copy and "forwarding" in interface_copy["vrf"]:
                vrf_name = interface_copy["vrf"]["forwarding"]
            else:
                vrf_name = "default"

            if not vrf_name in interfaces_by_vrf:
                interfaces_by_vrf[vrf_name] = []

            interfaces_by_vrf[vrf_name].append(interface_copy)

    return interfaces_by_vrf

def get_route_forwarding_list_by_vrf(config_before):
    route_forwarding_list_by_vrf = {}
    ip_obj = config_before.get("tailf-ned-cisco-ios:ip", {"route": {}}).get("route", {})

    route_forwarding_list_by_vrf["default"] = {
        common_xe.IP_FORWARDING_LIST: copy.deepcopy(ip_obj.get(common_xe.IP_FORWARDING_LIST, [])),
        common_xe.INTF_LIST: copy.deepcopy(ip_obj.get(common_xe.INTF_LIST, [])),
        common_xe.IP_INTF_FORWARDING_LIST: copy.deepcopy(ip_obj.get(common_xe.IP_INTF_FORWARDING_LIST, []))
    }

    for index, vrf in enumerate(ip_obj.get("vrf", [])):
        route_forwarding_list_by_vrf[vrf["name"]] = {
            "vrf-index": index,
            common_xe.IP_FORWARDING_LIST: copy.deepcopy(vrf.get(common_xe.IP_FORWARDING_LIST, [])),
            common_xe.INTF_LIST: copy.deepcopy(vrf.get(common_xe.INTF_LIST, [])),
            common_xe.IP_INTF_FORWARDING_LIST: copy.deepcopy(vrf.get(common_xe.IP_INTF_FORWARDING_LIST, []))
        }

    return route_forwarding_list_by_vrf

def configure_network_instances(config_before, config_leftover, interfaces_by_vrf, route_forwarding_list_by_vrf):
    for net_inst in openconfig_network_instances["openconfig-network-instance:network-instances"][
        "openconfig-network-instance:network-instance"]:
        configure_network_interfaces(net_inst, interfaces_by_vrf)

        if len(interfaces_by_vrf.get(net_inst["openconfig-network-instance:name"], [])) > 0:
            vrf_interfaces = interfaces_by_vrf.get(net_inst["openconfig-network-instance:name"])
            xe_ospfv2.configure_xe_ospf(net_inst, vrf_interfaces, config_before, config_leftover)
        if len(route_forwarding_list_by_vrf.get(net_inst["openconfig-network-instance:name"], [])) > 0:
            vrf_forwarding_list = route_forwarding_list_by_vrf.get(net_inst["openconfig-network-instance:name"])
            xe_static_route.configure_xe_static_routes(net_inst, vrf_forwarding_list, config_leftover, network_instances_notes)

def configure_network_interfaces(net_inst, interfaces_by_vrf):
    for interface in interfaces_by_vrf.get(net_inst["openconfig-network-instance:name"], []):
        name_split = interface["name"].split(".")
        primary_interface = name_split[0]
        new_interface = {
            "openconfig-network-instance:id": interface["type"] + interface["name"],
            "openconfig-network-instance:config": {
                "openconfig-network-instance:id": interface["type"] + interface["name"],
                "openconfig-network-instance:interface": interface["type"] + primary_interface
            }
        }

        if (interface["type"] != "Tunnel") and (interface["type"] != "Vlan") and (interface["type"] != "Port-channel"):
            subinterface = '0' if len(name_split) == 1 else name_split[1]
            new_interface["openconfig-network-instance:config"][
                "openconfig-network-instance:subinterface"] = subinterface
        elif interface["type"] == "Port-channel":  # Port-channel's don't have a sub-if 0
            if len(name_split) > 1:
                new_interface["openconfig-network-instance:config"]["openconfig-network-instance:subinterface"] = \
                name_split[1]

        net_inst["openconfig-network-instance:interfaces"]["openconfig-network-instance:interface"].append(new_interface)

def process_rd_rt(temp_vrf, vrf, vrf_index, config_leftover):
    if "rd" in vrf:
        temp_vrf["openconfig-network-instance:config"][
            "openconfig-network-instance:route-distinguisher"] = vrf["rd"]
        temp_vrf["openconfig-network-instance:config"][
            "openconfig-network-instance-ext:route-targets-import"] = []
        temp_vrf["openconfig-network-instance:config"][
            "openconfig-network-instance-ext:route-targets-export"] = []

        # RD is required to create RTs
        if "route-target" in vrf:
            process_rt(temp_vrf, vrf, "import")
            process_rt(temp_vrf, vrf, "export")
            del config_leftover["tailf-ned-cisco-ios:vrf"]["definition"][vrf_index]["route-target"]

        del config_leftover["tailf-ned-cisco-ios:vrf"]["definition"][vrf_index]["rd"]

def process_rt(temp_vrf, vrf, rt_type):
    for rt in vrf["route-target"].get(rt_type, []):
        if "asn-ip" in rt:
            temp_vrf["openconfig-network-instance:config"][
                f"openconfig-network-instance-ext:route-targets-{rt_type}"].append(rt["asn-ip"])

def cleanup_null_ospf_leftovers(config_leftover):
    ospf_leftover = config_leftover.get("tailf-ned-cisco-ios:router", {}).get("ospf", [])
    updated_ospf_list = []

    for ospf_index in range(len(ospf_leftover)):
        cleanup_neighbors(ospf_leftover[ospf_index])
        cleanup_traffic_area(ospf_leftover[ospf_index])
        cleanup_virtual_link(ospf_leftover[ospf_index])

        if len(ospf_leftover[ospf_index]) > 0:
            updated_ospf_list.append(ospf_leftover[ospf_index])
    
    if len(updated_ospf_list) > 0:
        config_leftover.get("tailf-ned-cisco-ios:router", {})["ospf"] = updated_ospf_list
    elif "ospf" in config_leftover.get("tailf-ned-cisco-ios:router", {}):
        del config_leftover["tailf-ned-cisco-ios:router"]["ospf"]

def cleanup_neighbors(ospf_leftover):
    if "neighbor" in ospf_leftover:
        del ospf_leftover["neighbor"]

def cleanup_virtual_link(ospf_leftover):
    if len(ospf_leftover.get("area", [])) < 1:
        return

    for area in ospf_leftover["area"]:
        updated_virtual_link_list = []

        for virtual_link in area.get("virtual-link", []):
            if virtual_link:
                updated_virtual_link_list.append(virtual_link)

        if len(updated_virtual_link_list) > 0:
            area["virtual-link"] = updated_virtual_link_list
        elif "virtual-link" in area:
            del area["virtual-link"]

def cleanup_traffic_area(ospf_leftover):
    if not "mpls" in ospf_leftover:
        return

    updated_traffic_area_list = []

    for area_item in ospf_leftover["mpls"].get("traffic-eng", {}).get("area", []):
        if area_item:
            updated_traffic_area_list.append(area_item)
        
    if len(updated_traffic_area_list) > 0:
        ospf_leftover["mpls"]["traffic-eng"]["area"] = updated_traffic_area_list
    elif "area" in ospf_leftover["mpls"].get("traffic-eng", {}):
        del ospf_leftover["mpls"]["traffic-eng"]["area"]

def cleanup_null_static_route_leftovers(config_leftover):
    if "route" in config_leftover.get("tailf-ned-cisco-ios:ip", {}):
        cleanup_static_routes(config_leftover["tailf-ned-cisco-ios:ip"]["route"])
    
    cleanup_vrf_null_leftover_static_routes(config_leftover)

    if "route" in config_leftover.get("tailf-ned-cisco-ios:ip", {}) and len(config_leftover["tailf-ned-cisco-ios:ip"]["route"]) == 0:
        del config_leftover["tailf-ned-cisco-ios:ip"]["route"]

def cleanup_vrf_null_leftover_static_routes(config_leftover):
    if len(config_leftover.get("tailf-ned-cisco-ios:ip", {"route": {}}).get("route", {}).get("vrf", [])) > 0:
        updated_vrf_list = []

        for vrf in config_leftover["tailf-ned-cisco-ios:ip"]["route"]["vrf"]:
            cleanup_static_routes(vrf)

            if len(vrf) > 0:
                updated_vrf_list.append(vrf)
        
        if len(updated_vrf_list) > 0:
            config_leftover["tailf-ned-cisco-ios:ip"]["route"]["vrf"] = updated_vrf_list
        else:
            del config_leftover["tailf-ned-cisco-ios:ip"]["route"]["vrf"]

def cleanup_static_routes(leftover_route):
    if common_xe.IP_FORWARDING_LIST in leftover_route:
        updated_ip_forwarding_list_leftover = get_updated_configs(leftover_route[common_xe.IP_FORWARDING_LIST])

        if len(updated_ip_forwarding_list_leftover) > 0:
            leftover_route[common_xe.IP_FORWARDING_LIST] = updated_ip_forwarding_list_leftover
        elif common_xe.IP_FORWARDING_LIST in leftover_route:
            del leftover_route[common_xe.IP_FORWARDING_LIST]
    if common_xe.INTF_LIST in leftover_route:
        updated_intf_list_leftover = get_updated_configs(leftover_route[common_xe.INTF_LIST])

        if len(updated_intf_list_leftover) > 0:
            leftover_route[common_xe.INTF_LIST] = updated_intf_list_leftover
        elif common_xe.INTF_LIST in leftover_route:
            del leftover_route[common_xe.INTF_LIST]
    if common_xe.IP_INTF_FORWARDING_LIST in leftover_route:
        updated_ip_intf_forwarding_list_leftover = get_updated_configs(leftover_route[common_xe.IP_INTF_FORWARDING_LIST])

        if len(updated_ip_intf_forwarding_list_leftover) > 0:
            leftover_route[common_xe.IP_INTF_FORWARDING_LIST] = updated_ip_intf_forwarding_list_leftover
        elif common_xe.IP_INTF_FORWARDING_LIST in leftover_route:
            del leftover_route[common_xe.IP_INTF_FORWARDING_LIST]
    if "name" in leftover_route and len(leftover_route) < 2:
        del leftover_route["name"]

def get_updated_configs(list_leftover):
    updated_static_list = []

    for item in list_leftover:
        if item:
            updated_static_list.append(item)

    return updated_static_list

def main(before: dict, leftover: dict, translation_notes: list = []) -> dict:
    """
    Translates NSO Device configurations to MDD OpenConfig configurations.

    Requires environment variables:
    NSO_URL: str
    NSO_USERNAME: str
    NSO_PASSWORD: str
    NSO_DEVICE: str
    TEST - If True, sends generated OC configuration to NSO Server: str

    :param before: Original NSO Device configuration: dict
    :param leftover: NSO Device configuration minus configs replaced with MDD OC: dict
    :return: MDD Openconfig Network Instances configuration: dict
    """

    xe_network_instances(before, leftover)
    translation_notes += network_instances_notes

    return openconfig_network_instances

if __name__ == "__main__":
    sys.path.append("../../")
    sys.path.append("../../../")

    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xe import common_xe
        from package_nso_to_oc.xe import xe_ospfv2
        from package_nso_to_oc.xe import xe_static_route
        from package_nso_to_oc import common
    else:
        import common_xe
        import xe_ospfv2
        import xe_static_route
        import common

    (config_before_dict, config_leftover_dict, interface_ip_dict) = common_xe.init_xe_configs()
    main(config_before_dict, config_leftover_dict)
    config_name = "_network_instances"
    config_remaining_name = "_remaining_network_instances"
    oc_name = "_openconfig_network_instances"
    common.print_and_test_configs(
        "xe1", config_before_dict, config_leftover_dict, openconfig_network_instances, 
        config_name, config_remaining_name, oc_name, network_instances_notes)
else:
    # This is needed for now due to top level __init__.py. We need to determine if contents in __init__.py is still necessary.
    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xe import common_xe
        from package_nso_to_oc.xe import xe_ospfv2
        from package_nso_to_oc.xe import xe_static_route
        from package_nso_to_oc import common
    else:
        from xe import common_xe
        from xe import xe_ospfv2
        from xe import xe_static_route
        import common



from setuptools import setup, find_namespace_packages
from pathlib import Path

includes = [
    "package_nso_to_oc",
    "package_nso_to_oc.*",
]

this_directory = Path(__file__).parent
long_description = (this_directory / "package_nso_to_oc" / "README.md").read_text()

setup(
    name='nso-oc',
    package_dir={'package_nso_to_oc': 'package_nso_to_oc'},
    packages=find_namespace_packages(
        include=includes
    ),
    description='Cisco NSO OpenConfig Tools',
    install_requires=[
        'urllib3',
    ],
    entry_points='''
        [console_scripts]
        nso-oc=package_nso_to_oc.main:main
    ''',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    url = 'https://github.com/model-driven-devops/nso-oc-services'
)